/***********************************************************
* Artsoft Retro-Game Library                               *
*----------------------------------------------------------*
*               Artsoft Entertainment                      *
*               Holger Schemel                             *
*               Detmolder Strasse 189                      *
*               33604 Bielefeld                            *
*               Germany                                    *
*               e-mail: info@artsoft.org                   *
*----------------------------------------------------------*
* zurl.c                                                   *
*----------------------------------------------------------*
* (c) by the owner of the alias/nickname "HerzAusGold"     *
*               e-mail: herzausgold@online.de              *
***********************************************************/
/*
 *
 * routines to get a file from the web 
 *----------------------------------------------------------
 * inspired from the examples on the curl homepage 
 * credits to:
 *   http://curl.haxx.se
 *   
*/

#include <stdio.h>
#include <stdlib.h>
#include <curl/curl.h>
#include <curl/types.h> /* new for v7 */
#include <curl/easy.h>  /* new for v7 */
#include "zurl.h"
#include "util.h"
#include "setup.h"
#include <string.h>

#define ZURL_QUOTE   '"'    /* " */
#define ZURL_SLASH   '/'    /* / */
#define ZURL_BSLASH  '\\'   /* \ */
#define ZURL_BLANK   ' '    /*   */
/*novalid in filename   \ / : * ? " < > |   **/
#define ZURL_DPOINT  ':'    /* : */
#define ZURL_ASTERIX '*'    /* * */
#define ZURL_QUEST   '?'    /* ? */  
#define ZURL_DQUOTE  '"'    /* " */
#define ZURL_LT      '<'    /* < */
#define ZURL_GT      '>'    /* > */
#define ZURL_PIPE    '|'    /* | */
#define ZURL_EOS     0      /* endOfString */


typedef double zurlBar_t;  // progessBar // only for illustration, debugging


/* myGetPath2 - build path from two  
*/  
static char *myGetPath2(char *path1, char *path2)
{
  char *complete_path = objectNew(strlen(path1) + 1 +
				  strlen(path2) + 1);
  sprintf(complete_path, "%s/%s", path1, path2);
  return complete_path;
}


/* writeFunction - write to file 
*/  
static int writeFunction(void *ptr, size_t size, size_t nmemb, zurl_t *zu)
{
  if (!zu->m_savefile) {
    if (zu->m_filename && zu->m_filename[0] != 0) {
        /* open */
        if (zu->m_fullfilename) {
          objectDelete((Pointer_t)&zu->m_fullfilename);
        }
        zu->m_fullfilename = myGetPath2(zu->m_pathname, zu->m_filename);
        zu->m_savefile = fopen(zu->m_fullfilename, "wb");
    } else { 
        // get filename from url
        char *pc = NULL;
        if ((pc = strrchr(zu->m_file_url, ZURL_SLASH))   || 
            (pc = strrchr(zu->m_file_url, ZURL_BSLASH))) {
          zu->m_filename = strNew(pc+1);
        } else {
          zu->m_filename = strNew(zu->m_file_url);
        }
        /* open */
        if (zu->m_filename && zu->m_filename[0] != 0) {
          if (zu->m_fullfilename) {
            objectDelete((Pointer_t)&zu->m_fullfilename);
          }
          zu->m_fullfilename = myGetPath2(zu->m_pathname, zu->m_filename);
          zu->m_savefile = fopen(zu->m_fullfilename, "wb");
        }
    }
  }
  if (zu->m_savefile) {
    return fwrite(ptr, size, nmemb, zu->m_savefile);
  } else {
    return 0;
  }
}


/* progressFunction - show the progress 
*/  
static int progressFunction(zurlBar_t *ptr,  // mydata
                            double t,   // dltotal 
                            double d,   // dlnow 
                            double ultotal,
                            double ulnow)
{
    *ptr = (double) (d*100.0/t);  // mydata
    fprintf(stderr, "%g / %g (%g %%)\n", d, t, d*100.0/t);
    return 0;
}


/* headerFunction - get header information
* rawBytes no trailing zero
*/
static size_t headerFunction( void *ptr, size_t itemsize, size_t nmemb, void *userptr)
{
    static char  seastr2[]  = {'f','i','l','e','n','a','m','e','=',ZURL_QUOTE,0};
    static int   lx2  = 10;
    char         *data    = ptr;
    size_t       fullsize = itemsize * nmemb;

    /* userptr */ 
    if (userptr != NULL) {
      zurl_header_t *hp = userptr;
      if (hp->m_mode & ZURL_FINDFILENAME) {
          size_t sz = 0;
          char *p2=NULL;
          char *pa=NULL, *pe=NULL;
          /* search */
          while (sz < fullsize) {
            if (sz < fullsize - lx2) {
              if (UTILstrNCaseCmp(data+sz, seastr2, lx2) == 0) {
                p2 = data + sz + lx2;
                sz += lx2;
                break;
              }
            }

            sz++;
          }

          /* p2 found */
          if (p2) {
            pa = p2;
            while (sz < fullsize) {
              if (data[sz] == ZURL_QUOTE) {
                pe = data+sz;
                break;
              }
              sz++;
            }
          }

          /* check */
          if (pa && pe && pe > pa) {
            char v = *pe;
            *pe = 0;
            if (*(hp->m_pfilename) != NULL) {
              objectDelete((Pointer_t)hp->m_pfilename);
            }
            *(hp->m_pfilename) = strNew(pa);
            *pe = v;
          } 
      } /* ZURL_FINDFILENAME */
    }

    return fullsize;
}


/* debugFunction - for strange debugging  
* rawBytes comes straight from libcurl and userptr from setopt DEBUGDATA, 
* though we're not using it here.
*/
static int debugFunction(CURL *handle,       /* the handle/transfer this concerns */
                         curl_infotype type, /* what kind of data */
                         char *data,         /* points to the data */
                         size_t size,        /* size of the data pointed to */
                         void *userptr)
{      
    size_t sz = size; 
    char  *dd = data;

    /* userptr */ 
    if (userptr != NULL) {
      zurl_debug_t *dp = userptr;
      if (dp->m_dbgfile != NULL) {
        fwrite (dd, sz, 1, dp->m_dbgfile);
      }
    }

    return 0;
}


/* checkFileSize - get file size 
*/  
static long checkFileSize(char* filename)
{
  long sz  = 0;
  FILE *fp = fopen(filename, "r");
  if (fp) {
    fseek(fp, 0, SEEK_END);
    sz = ftell(fp);
    fclose(fp);
  }
  return sz;
}


/*************************************************************
 zurl_easyGet - get and download file 
*************************************************************/
int zurl_easyGet(zurl_t *zu)
{    
    /* curl handle */
    CURL      *cuh;  // and they shout muh, muh...
    CURLcode  ret;
    double    filesize = 0;
    zurlBar_t bar = 0;   // progressBar 
 
    /* global init, only once in an application*/
    // curl_global_init(CURL_GLOBAL_ALL);

    /* init the curl session */
    cuh = curl_easy_init();
    if (cuh == NULL) {
      return CURLE_FAILED_INIT;  // if handle is null(no internet connection), abort file download
    }
     
    /*--- set easy options ---*/

    if (zu->m_flagFreshConnect) {
      curl_easy_setopt (cuh, CURLOPT_FRESH_CONNECT, TRUE);   //Set to use fresh connection every time
    }
    if (zu->m_optConnectTimeout < 0) {
      curl_easy_setopt (cuh, CURLOPT_CONNECTTIMEOUT, 5);     //Time out in second
    } else if (zu->m_optConnectTimeout > 0) {
      curl_easy_setopt (cuh, CURLOPT_CONNECTTIMEOUT, zu->m_optConnectTimeout);     
    }

    if (zu->m_filename != NULL && zu->m_filename[0] != 0) {
      zu->m_fullfilename = myGetPath2(zu->m_pathname, zu->m_filename);
      filesize = checkFileSize(zu->m_fullfilename);
      if (zu->m_resume && filesize > 0) {
        curl_easy_setopt (cuh, CURLOPT_RESUME_FROM, filesize);  //Resume file download start chunk from position XXX bytes
        zu->m_savefile = fopen (zu->m_fullfilename, "ab+");
      } else {
        zu->m_savefile = fopen (zu->m_fullfilename, "wb");
      }
    }
    curl_easy_setopt (cuh, CURLOPT_USERAGENT, "libcurl-agent/1.0");  // agent

    // some options (default == -1)
    if (zu->m_optBuffersize < 0) {
      curl_easy_setopt (cuh, CURLOPT_BUFFERSIZE, 500);      // File buffer size
    } else if (zu->m_optBuffersize > 0) {
      curl_easy_setopt (cuh, CURLOPT_BUFFERSIZE, zu->m_optBuffersize );      
    }
    if (zu->m_optLowSpeedLimit < 0) {
      curl_easy_setopt (cuh, CURLOPT_LOW_SPEED_LIMIT, 10);  // Time lower limit
    } else if (zu->m_optLowSpeedLimit > 0) {
      curl_easy_setopt (cuh, CURLOPT_LOW_SPEED_LIMIT, zu->m_optLowSpeedLimit);  
    }
    if (zu->m_optLowSpeedTime < 0) {
      curl_easy_setopt (cuh, CURLOPT_LOW_SPEED_TIME, 5);    // Download lower speed limit
    } else if (zu->m_optLowSpeedTime > 0) {
      curl_easy_setopt (cuh, CURLOPT_LOW_SPEED_TIME, zu->m_optLowSpeedTime);   
    }

    // proxy server
    if (zu->m_proxy_url != NULL && zu->m_proxy_url[0] != 0) {
      curl_easy_setopt (cuh, CURLOPT_PROXY, zu->m_proxy_url);  // using proxy server, format http://proxyserver:portno
    }

    // some flags
    if (zu->m_flagAutoReferer) {
      curl_easy_setopt (cuh, CURLOPT_AUTOREFERER, TRUE);    // Set to follow referer location
    }
    if (zu->m_flagFollowLocation) {
      curl_easy_setopt (cuh, CURLOPT_FOLLOWLOCATION, TRUE); // Set to follow location
    }

    // the one and only url
    curl_easy_setopt (cuh, CURLOPT_URL, zu->m_file_url);  // File download URL

    // post fields
    if (zu->m_use_post) {
      if (zu->m_post_fields != NULL && zu->m_post_fields[0] != 0) {
        curl_easy_setopt (cuh, CURLOPT_POST, 1);            // Set to 1 with POST field append
        curl_easy_setopt (cuh, CURLOPT_POSTFIELDS, zu->m_post_fields);  //Your field submit format "a=1&b=2&c=3"
      } else {
        curl_easy_setopt (cuh, CURLOPT_POST, 0);            // Set to 0 no POST field append
      }
    }
               
    // callback functions  // pointer to file-pointer (for findfile)                                          
    curl_easy_setopt (cuh, CURLOPT_WRITEDATA, zu);   // Get data here...
    curl_easy_setopt (cuh, CURLOPT_WRITEFUNCTION, writeFunction);
    if (zu->m_headerData != NULL) {
      curl_easy_setopt (cuh, CURLOPT_HEADERDATA, zu->m_headerData); // header info here...
      curl_easy_setopt (cuh, CURLOPT_HEADERFUNCTION, headerFunction);
    }
    if (zu->m_debugData != NULL) {
      curl_easy_setopt (cuh, CURLOPT_DEBUGDATA, zu->m_debugData); // and debug info here...
      curl_easy_setopt (cuh, CURLOPT_DEBUGFUNCTION, debugFunction);
    }
    
    // verbose, debug
    if (zu->m_verbose) {
      curl_easy_setopt (cuh, CURLOPT_VERBOSE, TRUE);
    }
    // progress
    if (zu->m_progress) {
      curl_easy_setopt (cuh, CURLOPT_NOPROGRESS, FALSE);      //Get file transfer progress here...
      curl_easy_setopt (cuh, CURLOPT_PROGRESSFUNCTION, progressFunction); 
      curl_easy_setopt (cuh, CURLOPT_PROGRESSDATA, (void*) &bar);      //Get progress data
    }

    // perform now!
    ret = curl_easy_perform(cuh);
    
    if (zu->m_savefile) {
      fclose(zu->m_savefile);
      zu->m_savefile = NULL;
    } 
    curl_easy_cleanup (cuh);
         
    /* global cleanup, only once in an application*/
    // curl_global_cleanup();

    return ret;   
} // zurl_easy_get


/*************************************************************
 zurl_get - get and download file 
*************************************************************/
int zurl_get(char *logdir, char *dir, char *url, 
             char *proxy, long debug,
             char **pFilename) 
{
  int               ret;
  zurl_t            zu;
  zurl_header_t     headerData;
  zurl_debug_t      debugData;
  char              *dbgfilename = NULL;

  zu.m_file_url          = url;    // pointer to url 
  zu.m_proxy_url         = proxy;  // proxy
  zu.m_use_post          = FALSE;  // no post 
  zu.m_post_fields       = NULL;   // no post fields
  zu.m_debug             = debug;

  headerData.m_mode        = ZURL_FINDFILENAME;
  headerData.m_pfilename   = &zu.m_filename;   // filled in zurl_easy_get

  debugData.m_dbgfile      = NULL;

  if (zu.m_debug) {
    dbgfilename = myGetPath2(logdir,"zdebug.tmp");
    debugData.m_dbgfile   = fopen(dbgfilename,"wb");
  }

  zu.m_optConnectTimeout  = 0;   // none
  zu.m_optBuffersize      = -1;  // default 
  zu.m_optLowSpeedLimit   = -1;  // default
  zu.m_optLowSpeedTime    = -1;  // default
  zu.m_flagFreshConnect   = 0;   // none
  zu.m_flagAutoReferer    = 1;   // on
  zu.m_flagFollowLocation = 1;   // on

  zu.m_resume             = 0;      // no resume
  zu.m_pathname           = dir;    // pathname
  zu.m_filename           = NULL;   // filename - header pointer here!
  zu.m_fullfilename       = NULL;   // full filename
  zu.m_savefile           = NULL;   // file pointer 

  zu.m_headerData         = &headerData;

  zu.m_debugData          = &debugData;
  zu.m_progress           = 0;   // progress=1,silent=0
  zu.m_verbose            = 1;   // verbose=1,silent=0

  ret = zurl_easyGet(&zu);

  if (zu.m_filename) {
    objectDelete((Pointer_t)&zu.m_filename);
  }
  if (zu.m_fullfilename) {
    if(pFilename) {
      *pFilename = strNew(zu.m_fullfilename);
    }
    objectDelete((Pointer_t)&zu.m_fullfilename);
  }
  if (zu.m_savefile) {
    fclose(zu.m_savefile);
    zu.m_savefile = NULL;
  }

  if (debugData.m_dbgfile) {
    fclose(debugData.m_dbgfile);
    debugData.m_dbgfile = NULL;
  }
  if (dbgfilename) {
    objectDelete((Pointer_t)&dbgfilename);
  }

  return ret;
}


/*************************************************************
 zurl_confget - check config - get and download file 
*************************************************************/
int zurl_confget(char *basedir, char *levdir, 
                 char *confurl, char **pFilename) 
{
  int  ret = 0;
  char *purl;
  char *p1;

  char *confFilename = NULL;
  char *realurl      = NULL;
  char *urlPrefix    = NULL;
  char *siteIdent    = NULL;
  char *proxy        = NULL;
  long debug         = 0;
  SetupFileHash *zurl_setup_hash;

  // get site ident
  purl = confurl;
  p1 = strchr(confurl, ':');
  if (p1) {
    *p1 = 0;
    siteIdent = strNew(confurl);
    strLower(siteIdent);
    purl = p1+1;
  }

  // get config, proxy, urlPrefix 
  confFilename = myGetPath2(basedir, ZURLSETUP_FILENAME);
  if ((zurl_setup_hash = loadSetupFileHash(confFilename)))
  {
    char *token_value;

    token_value = getHashEntry(zurl_setup_hash, ZURL_TOKEN_PROXY);
    if (token_value)
    {
      proxy = strNew(token_value);
    }
    token_value = getHashEntry(zurl_setup_hash, ZURL_TOKEN_DEBUG);
    if (token_value)
    {
      debug = atoi(token_value);
    }

    if (siteIdent != NULL) {
      token_value = getHashEntry(zurl_setup_hash, siteIdent);
      if (token_value)
      {
        urlPrefix = strNew(token_value);
      }
    }

    checkSetupFileHashIdentifier(zurl_setup_hash, confFilename, getCookie("ZURLSETUP"));

    freeSetupFileHash(zurl_setup_hash);
  }

  // if not found set default
  if (urlPrefix == NULL) {  // default
    if (atoi(purl) == 0) {
      urlPrefix = strNew("\0");  // no prefix
    } else {
      urlPrefix = strNew("http://www.zomis.net/rnd/download.php?id=");
    }
  }

  // make real url
  realurl = (char*) objectNew(strlen(urlPrefix)+strlen(purl)+1+1);
  if (realurl) { 
    if (strstr(urlPrefix, "%s")) {
      sprintf(realurl, urlPrefix, purl);
    } else {
      strcpy(realurl, urlPrefix);  
      strcat(realurl, purl);
    }  

    fprintf(stderr, "-I- zurl_get -- url: %s \n", realurl);
    ret = zurl_get(basedir, levdir, realurl, proxy, debug, pFilename);
    fprintf(stderr, "-I- zurl_get %d url: %s filename: %s\n", 
            ret, realurl, 
            (pFilename != NULL && *pFilename != NULL) ? *pFilename : "<none>");
  }
  objectDelete((Pointer_t)&realurl);
  objectDelete((Pointer_t)&confFilename);
  objectDelete((Pointer_t)&urlPrefix);
  objectDelete((Pointer_t)&siteIdent);
  objectDelete((Pointer_t)&proxy);

  return ret;
}

